#ifndef G4NCrystal_hh
#define G4NCrystal_hh

//////////////////////////////////////////////////////
// Convenience header for all G4NCrystal interfaces //
//////////////////////////////////////////////////////

#ifndef G4NCrystal_Install_hh
#  include "G4NCrystal/G4NCInstall.hh"
#endif
#ifndef G4NCrystal_MatHelper_hh
#  include "G4NCrystal/G4NCMatHelper.hh"
#endif
#ifndef G4NCrystal_Bias_hh
#  include "G4NCrystal/G4NCBias.hh"
#endif

#endif
